import importlib
from unittest.mock import patch

import pytest

import outlines.models as models
from outlines.generate.api import GenerationParameters, SamplingParameters
from outlines.models.exllamav2 import ExLlamaV2Model


@pytest.fixture(scope="session")
def model_exllamav2(tmp_path_factory):
    return models.exl2(
        model_path="blockblockblock/TinyLlama-1.1B-Chat-v1.0-bpw4-exl2",
        cache_q4=True,
        paged=False,
    )


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_exl2_import_error(request, model_fixture):
    with patch.dict("sys.modules", {"exllamav2": None}):
        with pytest.raises(ImportError):
            models.exl2(
                model_path="blockblockblock/TinyLlama-1.1B-Chat-v1.0-bpw4-exl2",
                cache_q4=True,
                paged=False,
            )


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_model_attributes(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    assert hasattr(model, "generator")
    assert hasattr(model, "tokenizer")
    assert model.tokenizer.convert_token_to_string(1) == 1
    assert hasattr(model, "max_seq_len")
    assert isinstance(model.max_seq_len, int)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_model_generate_prompt_types(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    prompt = "test"
    generation_params = GenerationParameters(max_tokens=10, stop_at=None, seed=None)
    structure_logits_processor = None
    sampling_params = SamplingParameters(
        "multinomial",
        1,
        0.9,
        50,
        1.0,
    )
    output = model.generate(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    assert isinstance(output, str)
    prompt = ["test"]
    output = model.generate(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    assert isinstance(output, str)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_model_generate_no_max_tokens(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    prompt = "test"
    generation_params = GenerationParameters(max_tokens=None, stop_at=None, seed=None)
    structure_logits_processor = None
    sampling_params = SamplingParameters(
        "multinomial",
        1,
        0.9,
        50,
        1.0,
    )
    output = model.generate(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    assert isinstance(output, str)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_model_generate_test_stop_at(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    prompt = "test"
    generation_params = GenerationParameters(max_tokens=10, stop_at="stop", seed=None)
    structure_logits_processor = None
    sampling_params = SamplingParameters(
        "multinomial",
        1,
        0.9,
        50,
        1.0,
    )
    output = model.generate(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    assert isinstance(output, str)
    generation_params = GenerationParameters(max_tokens=10, stop_at=["stop"], seed=None)
    output = model.generate(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    assert isinstance(output, str)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_model_generate_multisampling(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    prompt = "test"
    generation_params = GenerationParameters(max_tokens=10, stop_at="stop", seed=None)
    structure_logits_processor = None
    sampling_params = SamplingParameters(
        "multinomial",
        2,
    )
    output = model.generate(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    assert isinstance(output, list)
    assert isinstance(output[0], str)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_model_prepare_generation_parameters(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    prompt = "test"
    generation_params = GenerationParameters(max_tokens=10, stop_at="stop", seed=None)
    structure_logits_processor = None
    sampling_params = SamplingParameters(
        "multinomial",
        2,
    )
    exllamav2_params, prompts = model.prepare_generation_parameters(
        prompt, generation_params, sampling_params, structure_logits_processor
    )
    assert isinstance(exllamav2_params, dict)
    assert isinstance(prompts, list)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_model_stream_prompt_types(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    prompt = "test"
    generation_params = GenerationParameters(max_tokens=10, stop_at=None, seed=None)
    structure_logits_processor = None
    sampling_params = SamplingParameters(
        "multinomial",
        1,
        0.9,
        50,
        1.0,
    )
    generator = model.stream(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    for token in generator:
        assert isinstance(token, str)
    prompt = ["test"]
    generator = model.stream(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    for token in generator:
        assert isinstance(token, str)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_model_stream_no_max_tokens(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    prompt = "test"
    generation_params = GenerationParameters(max_tokens=None, stop_at=None, seed=None)
    structure_logits_processor = None
    sampling_params = SamplingParameters(
        "multinomial",
        1,
        0.9,
        50,
        1.0,
    )
    generator = model.stream(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    for token in generator:
        assert isinstance(token, str)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_model_stream_test_stop_at(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    prompt = "test"
    generation_params = GenerationParameters(max_tokens=10, stop_at="stop", seed=None)
    structure_logits_processor = None
    sampling_params = SamplingParameters(
        "multinomial",
        1,
        0.9,
        50,
        1.0,
    )
    generator = model.stream(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    for token in generator:
        assert isinstance(token, str)
    generation_params = GenerationParameters(max_tokens=10, stop_at=["stop"], seed=None)
    generator = model.stream(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    for token in generator:
        assert isinstance(token, str)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_model_stream_multisampling(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    prompt = "test"
    generation_params = GenerationParameters(max_tokens=10, stop_at="stop", seed=None)
    structure_logits_processor = None
    sampling_params = SamplingParameters(
        "multinomial",
        2,
    )
    generator = model.stream(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    for token in generator:
        assert isinstance(token, list)
        assert isinstance(token[0], str)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_model_stream_seed(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    prompt = "test"
    generation_params = GenerationParameters(max_tokens=10, seed=1, stop_at=None)
    structure_logits_processor = None
    sampling_params = SamplingParameters(
        "multinomial",
        1,
        0.9,
        50,
        1.0,
    )
    generator = model.stream(
        prompt, generation_params, structure_logits_processor, sampling_params
    )
    for token in generator:
        assert isinstance(token, str)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_reformat_output(request, model_fixture):
    model = request.getfixturevalue(model_fixture)
    sampling_params = SamplingParameters(
        "multinomial",
        1,
    )
    output = "test"
    reformatted_output = model.reformat_output(output, sampling_params)
    assert reformatted_output == output
    output = ["test"]
    reformatted_output = model.reformat_output(output, sampling_params)
    assert reformatted_output == output[0]
    output = ["test", "test"]
    sampling_params = SamplingParameters(
        "multinomial",
        1,
    )
    reformatted_output = model.reformat_output(output, sampling_params)
    assert len(reformatted_output) == 2
    assert reformatted_output[0] == "test"
    assert reformatted_output[1] == "test"
    output = ["test", "test"]
    sampling_params = SamplingParameters(
        "multinomial",
        2,
    )
    reformatted_output = model.reformat_output(output, sampling_params)
    assert len(reformatted_output) == 2
    assert reformatted_output[0] == "test"
    assert reformatted_output[1] == "test"
    output = ["test", "test", "test", "test"]
    sampling_params = SamplingParameters(
        "multinomial",
        2,
    )
    reformatted_output = model.reformat_output(output, sampling_params)
    assert len(reformatted_output) == 2
    assert reformatted_output[0] == ["test", "test"]
    assert reformatted_output[1] == ["test", "test"]


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_exl2_max_chunk_size(request, model_fixture):
    model = models.exl2(
        model_path="blockblockblock/TinyLlama-1.1B-Chat-v1.0-bpw4-exl2",
        cache_q4=True,
        paged=False,
        max_chunk_size=128,
    )
    assert isinstance(model, ExLlamaV2Model)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_exl2_cache_default(request, model_fixture):
    model = models.exl2(
        model_path="blockblockblock/TinyLlama-1.1B-Chat-v1.0-bpw4-exl2",
        paged=False,
    )
    assert isinstance(model, ExLlamaV2Model)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def is_flash_attn_available():
    try:
        importlib.import_module("flash_attn")
    except (ImportError, AssertionError):
        return False
    return True


@pytest.mark.skipif(not is_flash_attn_available(), reason="flash-attn is not installed")
@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_exl2_paged(request, model_fixture):
    model = models.exl2(
        model_path="blockblockblock/TinyLlama-1.1B-Chat-v1.0-bpw4-exl2",
        cache_q4=True,
        paged=True,
    )
    assert isinstance(model, ExLlamaV2Model)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_exl2_draft_model(request, model_fixture):
    model = models.exl2(
        model_path="blockblockblock/TinyLlama-1.1B-Chat-v1.0-bpw4-exl2",
        draft_model_path="blockblockblock/TinyLlama-1.1B-Chat-v1.0-bpw4-exl2",
        cache_q4=True,
        paged=False,
    )
    assert isinstance(model, ExLlamaV2Model)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_exl2_draft_model_cache_default(request, model_fixture):
    model = models.exl2(
        model_path="blockblockblock/TinyLlama-1.1B-Chat-v1.0-bpw4-exl2",
        draft_model_path="blockblockblock/TinyLlama-1.1B-Chat-v1.0-bpw4-exl2",
        paged=False,
    )
    assert isinstance(model, ExLlamaV2Model)


@pytest.mark.parametrize("model_fixture", ["model_exllamav2"])
def test_exl2_set_max_seq_len(request, model_fixture):
    model = models.exl2(
        model_path="blockblockblock/TinyLlama-1.1B-Chat-v1.0-bpw4-exl2",
        max_seq_len=2048,
        paged=False,
        cache_q4=True,
    )
    assert isinstance(model, ExLlamaV2Model)



import pytest

from outlines.models.mlxlm import mlxlm
from outlines.models.transformers import TransformerTokenizer

try:
    import mlx.core as mx

    HAS_MLX = mx.metal.is_available()
except ImportError:
    HAS_MLX = False


TEST_MODEL = "mlx-community/SmolLM-135M-Instruct-4bit"


@pytest.mark.skipif(not HAS_MLX, reason="MLX tests require Apple Silicon")
def test_mlxlm_model():
    model = mlxlm(TEST_MODEL)
    assert hasattr(model, "model")
    assert hasattr(model, "tokenizer")
    assert isinstance(model.tokenizer, TransformerTokenizer)


@pytest.mark.skipif(not HAS_MLX, reason="MLX tests require Apple Silicon")
def test_mlxlm_tokenizer():
    model = mlxlm(TEST_MODEL)

    # Test single string encoding/decoding
    test_text = "Hello, world!"
    token_ids = mx.array(model.mlx_tokenizer.encode(test_text))
    assert isinstance(token_ids, mx.array)


@pytest.mark.skipif(not HAS_MLX, reason="MLX tests require Apple Silicon")
def test_mlxlm_generate():
    from outlines.generate.api import GenerationParameters, SamplingParameters

    model = mlxlm(TEST_MODEL)
    prompt = "Write a haiku about programming:"

    # Test with basic generation parameters
    gen_params = GenerationParameters(max_tokens=50, stop_at=None, seed=None)

    # Test with different sampling parameters
    sampling_params = SamplingParameters(
        sampler="multinomial", num_samples=1, top_p=0.9, top_k=None, temperature=0.7
    )

    # Test generation
    output = model.generate(prompt, gen_params, None, sampling_params)
    assert isinstance(output, str)
    assert len(output) > 0


@pytest.mark.skipif(not HAS_MLX, reason="MLX tests require Apple Silicon")
def test_mlxlm_stream():
    from outlines.generate.api import GenerationParameters, SamplingParameters

    model = mlxlm(TEST_MODEL)
    prompt = "Count from 1 to 5:"

    gen_params = GenerationParameters(max_tokens=20, stop_at=None, seed=None)

    sampling_params = SamplingParameters(
        sampler="greedy",  # Use greedy sampling for deterministic output
        num_samples=1,
        top_p=None,
        top_k=None,
        temperature=0.0,
    )

    # Test streaming
    stream = model.stream(prompt, gen_params, None, sampling_params)
    tokens = list(stream)
    assert len(tokens) > 0
    assert all(isinstance(token, str) for token in tokens)

    # Test that concatenated streaming output matches generate output
    streamed_text = "".join(tokens)
    generated_text = model.generate(prompt, gen_params, None, sampling_params)
    assert streamed_text == generated_text


@pytest.mark.skipif(not HAS_MLX, reason="MLX tests require Apple Silicon")
def test_mlxlm_errors():
    model = mlxlm(TEST_MODEL)

    # Test batch inference (should raise NotImplementedError)
    with pytest.raises(NotImplementedError):
        from outlines.generate.api import GenerationParameters, SamplingParameters

        gen_params = GenerationParameters(max_tokens=10, stop_at=None, seed=None)
        sampling_params = SamplingParameters("multinomial", 1, None, None, 1.0)
        model.generate(["prompt1", "prompt2"], gen_params, None, sampling_params)

    # Test beam search (should raise NotImplementedError)
    with pytest.raises(NotImplementedError):
        sampling_params = SamplingParameters("beam_search", 1, None, None, 1.0)
        model.generate("test prompt", gen_params, None, sampling_params)



from .api import SequenceGenerator
from .cfg import cfg
from .choice import choice
from .format import format
from .fsm import fsm
from .json import json
from .regex import regex
from .text import text
